local _G = _G
local date = _G.date
local MAJOR_VERSION = 'GraveLevelBar'

local LibStub = _G.LibStub
local AceAddon = LibStub('AceAddon-3.0')
local addon = AceAddon:GetAddon(MAJOR_VERSION)

local min = min

addon.xpBarTemplate = {}
local xpBarTemplate = addon.xpBarTemplate

xpBarTemplate.defaults = {
    startXP = 0,
    targetXP = nil,
}

xpBarTemplate.optionsTable = {
    name = nil,
    type = 'group',
    args = {
        startXP = {
            name = "Start XP",
            desc = "Minimum XP value to display.",
            type = 'range',
            min = 0,
            max = nil,
            step = 1,
            bigStep = nil,
            get = 'getStartXP',
            set = 'setStartXP',
        },
        startLevel = {
            name = "Start Level",
            desc = "Minimum level to display.",
            type = 'range',
            min = 1,
            max = nil,
            bigStep = 1,
            get = 'getStartLevel',
            set = 'setStartLevel',
        },
        targetXP = {
            name = "Target XP",
            desc = "Maximum XP value to disply.",
            type = 'range',
            min = nil,
            max = nil,
            step = 1,
            bigStep = nil,
            get = 'getTargetXP',
            set = 'setTargetXP',
        },
        targetLevel = {
            name = "Target Level",
            desc = "Maximum level to disply.",
            type = 'range',
            min = nil,
            max = nil,
            bigStep = 1,
            get = 'getTargetLevel',
            set = 'setTargetLevel',
        },
    },
}

-- start xp
function xpBarTemplate:getStartXP(info)
    local startXP = self:getOptions().startXP
    local startXPMax = self:getStartXPMax()
    if startXP > self:getStartXPMax() then
        return startXPMax
    end
    return self:getOptions().startXP
end

function xpBarTemplate:getStartXPMin()
    error("Implement me")
end

function xpBarTemplate:getStartXPMax()
    return self:getCurrentXP()
end

function xpBarTemplate:setStartXP(info,value)
    local startXPMin = self:getStartXPMin()
    local startXPMax = self:getStartXPMax()
    if value < startXPMin then
        value = startXPMin
    end
    if value > startXPMax then
        value = startXPMax
    end
    self:getOptions().startXP = value
    self:Invalidate()
end

-- start level
function xpBarTemplate:getStartLevel(info)
    local startXP = self:getStartXP()
    local level = self:calcLevelForXP(startXP)
    return level
end

function xpBarTemplate:getStartLevelMin()
    local startXPMin = self:getStartXPMin()
    local levelMin = self:calcLevelForXP(startXPMin)
    return levelMin
end

function xpBarTemplate:getStartLevelMax()
    local startXPMax = self:getStartXPMax()
    local levelMax = self:calcLevelForXP(startXPMax)
    return levelMax
end

function xpBarTemplate:setStartLevel(info,value)
    local prevLevel = floor(value)
    local prevLevelXP = self:calcXpForLevel(prevLevel)
    local nextLevel = ceil(value)
    if prevLevel < nextLevel then
        local nextLevelXP = self:calcXpForLevel(nextLevel)
        local startXP =  prevLevelXP + (value-prevLevel)*(nextLevelXP-prevLevelXP)
        self:setStartXP(info,startXP)
    else
        self:setStartXP(info,prevLevelXP)
    end
end

-- target xp
function xpBarTemplate:getTargetXP(info)
    return self:getOptions().targetXP or self:getTargetXPMax()
end

function xpBarTemplate:getTargetXPMin()
    return self:getCurrentXP()
end

function xpBarTemplate:getTargetXPMax()
    local maxPlayerLevel = self:getTargetLevelMax()
    return self:calcXpForLevel(maxPlayerLevel)
end

function xpBarTemplate:setTargetXP(info,value)
    self:getOptions().targetXP = value
    self:Invalidate()
end

-- target level
function xpBarTemplate:getTargetLevel(info)
    local targetXP = self:getTargetXP()
    local level = self:calcLevelForXP(targetXP)
    return level
end

function xpBarTemplate:getTargetLevelMin()
    local targetXPMin = self:getTargetXPMin()
    local levelMin = self:calcLevelForXP(targetXPMin)
    return levelMin
end

function xpBarTemplate:getTargetLevelMax()
    local targetXPMax = self:getTargetXPMax()
    local levelMax = self:calcLevelForXP(targetXPMax)
    return levelMax
end

function xpBarTemplate:OnInitialize()
    self.superOnInitialize(self)
    self:updateOptionsTable()
end

function xpBarTemplate:OnEnable()
    self.superOnEnable(self)
    self:updateOptionsTable()
end

function xpBarTemplate:setTargetLevel(info,value)
    assert(self and self.optionsTable and self.optionsTable.name, "where is my optionsTable?")
    assert(value and type(value) == 'number',"suplied argument value is not a number: name=" .. tostring(self.optionsTable.name) .. ", value=" .. tostring(value))
    local prevLevel = floor(value)
    local prevLevelXP = self:calcXpForLevel(prevLevel)
    local nextLevel = ceil(value)
    assert(prevLevel,"prev level is nil. name=" .. tostring(self.optionsTable.name) .. " value=" .. tostring(value))
    assert(nextLevel,"next level is nil. name=" .. tostring(self.optionsTable.name) .. " value=" .. tostring(value))
    if prevLevel < nextLevel then
        local nextLevelXP = self:calcXpForLevel(nextLevel)
        local targetXP =  prevLevelXP + (value-prevLevel)*(nextLevelXP-prevLevelXP)
        self:setTargetXP(info,targetXP)
    else
        self:setTargetXP(info,prevLevelXP)
    end
end

-- end of setters and getters


function xpBarTemplate:updateOptionsTable()
    --update options table
    local startXPMin = self:getStartXPMin()
    local startXPMax = self:getStartXPMax()
    self.optionsTable.args.startXP.min = startXPMin
    self.optionsTable.args.startXP.max = startXPMax
    self.optionsTable.args.startLevel.max = self:getStartLevelMax()
    self.optionsTable.args.targetXP.max = self:getTargetXPMax()
    self.optionsTable.args.targetLevel.max = self:getTargetLevelMax()
    self.optionsTable.args.targetXP.min = self.optionsTable.args.startXP.max
    self.optionsTable.args.targetLevel.min = self.optionsTable.args.startLevel.max
    assert(startXPMin <= startXPMax, "startXPMin=" .. tostring(startXPMin) .. ", startXPMax=" .. tostring(startXPMax))
    assert(self.optionsTable.args.startLevel.min  <= self.optionsTable.args.startLevel.max)
    assert(self.optionsTable.args.targetXP.min    <= self.optionsTable.args.targetXP.max)
    assert(self.optionsTable.args.targetLevel.min <= self.optionsTable.args.targetLevel.max)
end

function xpBarTemplate:getCurrentXP()
    error("Not implemented")
end

function addon:createXPModule(name,...)
    assert(addon.barTemplate.optionsTable.type)
    local template = addon:merge(addon.barTemplate, xpBarTemplate)
    assert(template.optionsTable.type)
    module = addon:NewModule(name,template,'AceHook-3.0',...)
    return module
end
